from functools import cached_property, reduce
from typing import Optional

import sympy
from sympy.physics.paulialgebra import Pauli, evaluate_pauli_product

from qibo import gates
from qibo.backends import Backend, _check_backend
from qibo.config import raise_error
from qibo.symbols import I, X, Y, Z

# For mapping between qibo symbols and sympy Pauli objects
PAULI_MAPPING = {X: Pauli(1), Y: Pauli(2), Z: Pauli(3)}
INVERSE_PAULI_MAPPING = {_v: _k for _k, _v in PAULI_MAPPING.items()}


class HamiltonianTerm:
    """Term of a :class:`qibo.hamiltonians.hamiltonians.SymbolicHamiltonian`.

    Symbolic Hamiltonians are represented by a list of
    :class:`qibo.hamiltonians.terms.HamiltonianTerm` objects storred in the
    ``SymbolicHamiltonian.terms`` attribute. The mathematical expression of
    the Hamiltonian is the sum of these terms.

    Args:
        matrix (np.ndarray): Full matrix corresponding to the term representation
            in the computational basis. Has size (2^n, 2^n) where n is the
            number of target qubits of this term.
        q (list): List of target qubit ids.
    """

    def __init__(self, matrix, *q, backend: Optional[Backend] = None):
        self.backend = _check_backend(backend)
        for qi in q:
            if qi < 0:
                raise_error(
                    ValueError,
                    f"Invalid qubit id {qi} < 0 was given in Hamiltonian term.",
                )
        if not isinstance(matrix, self.backend.tensor_types):
            raise_error(TypeError, f"Invalid type {type(matrix)} of symbol matrix.")
        dim = int(matrix.shape[0])
        if 2 ** len(q) != dim:
            raise_error(
                ValueError,
                f"Matrix dimension {dim} given in Hamiltonian "
                + "term is not compatible with the number "
                + f"of target qubits {len(q)}.",
            )
        self.target_qubits = tuple(q)
        self._gate = None
        self.hamiltonian = None
        self._matrix = self.backend.cast(matrix)

    @property
    def matrix(self):
        """Matrix representation of the term."""
        return self._matrix

    @property
    def gate(self):
        """:class:`qibo.gates.gates.Unitary` gate that implements the action of the term on states."""
        if self._gate is None:
            self._gate = gates.Unitary(self.matrix, *self.target_qubits)
        return self._gate

    def exp(self, x):
        """Matrix exponentiation of the term."""
        return self.backend.calculate_matrix_exp(self.matrix, phase=-1j * x)

    def expgate(self, x):
        """:class:`qibo.gates.gates.Unitary` gate implementing the action of exp(term) on states."""
        return gates.Unitary(self.exp(x), *self.target_qubits)

    def merge(self, term):
        """Creates a new term by merging the given term to the current one.

        The resulting term corresponds to the sum of the two original terms.
        The target qubits of the given term should be a subset of the target
        qubits of the current term.
        """
        if not set(term.target_qubits).issubset(set(self.target_qubits)):
            raise_error(
                ValueError,
                "Cannot merge HamiltonianTerm acting on "
                + f"qubits {term.target_qubits} to term on qubits {self.target_qubits}.",
            )
        matrix = self.backend.np.kron(
            term.matrix, self.backend.matrices.I(2 ** (len(self) - len(term)))
        )
        matrix = self.backend.np.reshape(matrix, 2 * len(self) * (2,))
        order = []
        i = len(term)
        for qubit in self.target_qubits:
            if qubit in term.target_qubits:
                order.append(term.target_qubits.index(qubit))
            else:
                order.append(i)
                i += 1
        order.extend([x + len(order) for x in order])
        matrix = self.backend.np.transpose(matrix, order)
        matrix = self.backend.np.reshape(matrix, 2 * (2 ** len(self),))
        return HamiltonianTerm(
            self.matrix + matrix, *self.target_qubits, backend=self.backend
        )

    def __len__(self):
        return len(self.target_qubits)

    def __mul__(self, x):
        return HamiltonianTerm(
            x * self.matrix, *self.target_qubits, backend=self.backend
        )

    def __rmul__(self, x):
        return self.__mul__(x)

    def __call__(self, backend, state, nqubits, gate=None, density_matrix=False):
        """Applies the term on a given state vector or density matrix."""
        # TODO: improve this and understand why it works
        if isinstance(gate, bool) or gate is None:
            gate = self.gate
        if density_matrix:
            return backend.apply_gate_half_density_matrix(gate, state, nqubits)
        return backend.apply_gate(gate, state, nqubits)  # pylint: disable=E1102


class SymbolicTerm(HamiltonianTerm):
    """:class:`qibo.hamiltonians.terms.HamiltonianTerm` constructed using ``sympy`` expression.

    Example:
        .. testcode::

            from qibo.symbols import X, Y
            from qibo.hamiltonians.terms import SymbolicTerm
            sham = X(0) * X(1) + 2 * Y(0) * Y(1)
            termsdict = sham.as_coefficients_dict()
            sterms = [SymbolicTerm(c, f) for f, c in termsdict.items()]

    Args:
        coefficient (complex): Complex number coefficient of the underlying
            term in the Hamiltonian.
        factors (sympy.Expr): Sympy expression for the underlying term.
    """

    def __init__(self, coefficient, factors=1, backend: Optional[Backend] = None):
        self._gate = None
        self.hamiltonian = None
        self.backend = _check_backend(backend)
        self.coefficient = complex(coefficient)

        # List of :class:`qibo.symbols.Symbol` that represent the term factors
        self.factors = []
        # Dictionary that maps target qubit ids to a list of matrices that act on each qubit
        self.matrix_map = {}
        if factors != 1:
            for factor in factors.as_ordered_factors():
                factor, power = self._compute_power(factor)
                if isinstance(factor, sympy.Symbol):
                    # Skip any qibo.symbols.I factors
                    if isinstance(factor, I):
                        continue
                    # forces the backend of the factor
                    # this way it is not necessary to explicitely define the
                    # backend of a symbol, i.e. Z(q, backend=backend)
                    factor.backend = self.backend
                    if isinstance(factor.matrix, self.backend.tensor_types):
                        q = factor.target_qubit
                        if q in self.matrix_map:
                            # Check for possible simplifications only if current factor is X/Y/Z
                            if isinstance(factor, (X, Y, Z)):
                                self._simplify_q_factors(factor, q)
                            # Case whereby factor isn't X/Y/Z: Multiply matrix multiple times based on its power,
                            # see the ``SymbolicTerm.matrix`` property for the full matrix calculation
                            else:
                                self.factors.extend(power * [factor])
                                self.matrix_map[q].extend(power * [factor.matrix])
                        else:
                            self.factors.extend(power * [factor])
                            self.matrix_map[q] = power * [factor.matrix]

                    else:
                        self.coefficient *= factor.matrix
                elif factor == sympy.I:
                    self.coefficient *= 1j
                elif factor.is_number:
                    self.coefficient *= complex(factor)
                else:  # pragma: no cover
                    raise_error(TypeError, f"Cannot parse factor {factor}.")

        self.target_qubits = tuple(sorted(self.matrix_map.keys()))

    @staticmethod
    def _compute_power(factor):
        """Break down a factor into a base term and its power, e.g. `X**n` returns `X` and `n`. In case the base term is a Pauli also simplify the power calculation, i.e. return the identity for even `n` and the Pauli itself for odd `n`."""
        if isinstance(factor, sympy.Pow):
            factor, power = factor.args
            assert isinstance(power, sympy.Integer)
            assert isinstance(factor, sympy.Symbol)
            # if the symbol is a Pauli (i.e. a qibo symbol) and `power` is even
            # the power is the identity, thus the factor vanishes. Otherwise,
            # for an odd exponent, it remains unchanged (i.e. `power`=1)
            if factor.__class__ in (I, X, Y, Z):
                if not int(power) % 2:
                    factor = sympy.N(1)
                else:
                    power = 1
            else:
                power = int(power)
        else:
            power = 1
        return factor, power

    def _simplify_q_factors(self, factor, q):
        """Simplify the factors acting on qubit ``q`` using Pauli algebra after factor is added to ``q``."""
        q_factors = [_factor for _factor in self.factors if _factor.target_qubit == q]
        while q_factors:
            q_factor = q_factors[-1]
            # Only simplify if the last term in self.matrix_map[q] is a Pauli matrix
            if q_factor.__class__ in PAULI_MAPPING:
                pauli_product = evaluate_pauli_product(
                    PAULI_MAPPING[q_factor.__class__] * PAULI_MAPPING[factor.__class__]
                )
                # Move -1 or sympy.I from the product into the coefficient
                if -1 in pauli_product.atoms():
                    pauli_product *= -1
                    self.coefficient *= -1
                if sympy.I in pauli_product.atoms():
                    pauli_product *= -sympy.I
                    self.coefficient *= 1j
                # If pauli_product is not identity, then update the original factor/matrix
                factor_index = (
                    len(self.factors) - self.factors[::-1].index(q_factor) - 1
                )
                if pauli_product in INVERSE_PAULI_MAPPING:
                    factor = INVERSE_PAULI_MAPPING[pauli_product](q)
                    factor.backend = self.backend
                    self.factors[factor_index] = factor
                    self.matrix_map[q][-1] = factor.matrix
                # Otherwise, just remove the factor/matrix
                else:
                    self.factors.pop(factor_index)
                    self.matrix_map[q].pop(-1)
                    if not self.matrix_map[q]:
                        del self.matrix_map[q]
                q_factors.pop(-1)
            # X/Y/Z factor wasn't the last term previously => Cannot simplify, just extend
            else:
                self.factors.extend([factor])
                self.matrix_map[q].extend([factor.matrix])
                break

    @cached_property
    def matrix(self):
        """Calculates the full matrix corresponding to this term.

        Returns:
            Matrix as a ``np.ndarray`` of shape ``(2 ** ntargets, 2 ** ntargets)``
            where ``ntargets`` is the number of qubits included in the factors
            of this term.
        """
        matrices = list(self.qubit_to_matrix_map.values())
        return complex(self.coefficient) * reduce(self.backend.np.kron, matrices)

    @cached_property
    def qubit_to_matrix_map(self) -> dict:
        """Dictionary mapping each qubit to the corresponding global matrix
        that acts on it, i.e. the product of the matrices of all the factors
        acting on it.
        """
        return {
            q: reduce(self.backend.np.matmul, self.matrix_map.get(q))
            for q in self.target_qubits
        }

    def copy(self):
        """Creates a shallow copy of the term with the same attributes."""
        new = self.__class__(self.coefficient)
        new.factors = self.factors
        new.matrix_map = self.matrix_map
        new.target_qubits = self.target_qubits
        new.backend = self.backend
        return new

    def __mul__(self, x):
        """Multiplication of scalar to the Hamiltonian term."""
        new = self.copy()
        new.coefficient *= x
        return new

    def __call__(self, backend, state, nqubits, density_matrix=False):
        for factor in self.factors:
            state = super().__call__(
                backend, state, nqubits, factor.gate, density_matrix
            )
        return self.coefficient * state

    def commute(self, term) -> bool:
        """Check whether this term commutes with another term."""
        for q in set(self.target_qubits).intersection(set(term.target_qubits)):
            m1 = self.qubit_to_matrix_map.get(q)
            m2 = term.qubit_to_matrix_map.get(q)
            if not self.backend.np.all(m1 @ m2 == m2 @ m1):
                return False
        return True


class TermGroup(list):
    """Collection of multiple :class:`qibo.hamiltonians.terms.HamiltonianTerm` objects.

    Allows merging multiple terms to a single one for faster exponentiation
    during Trotterized evolution.

    Args:
        term (:class:`qibo.hamiltonians.terms.HamiltonianTerm`): Parent term of the group.
            All terms appended later should target a subset of the parents'
            target qubits.
    """

    def __init__(self, term):
        super().__init__([term])
        self.target_qubits = set(term.target_qubits)
        self._term = None

    def append(self, term):
        """Appends a new :class:`qibo.hamiltonians.terms.HamiltonianTerm` to the collection."""
        super().append(term)
        self.target_qubits |= set(term.target_qubits)
        self._term = None

    def can_append(self, term):
        """Checks if a term can be appended to the group based on its target qubits."""
        return set(term.target_qubits).issubset(self.target_qubits)

    @classmethod
    def from_terms(cls, terms):
        """Divides a list of terms to multiple :class:`qibo.hamiltonians.terms.TermGroup`s.

        Terms that target the same qubits are grouped to the same group.

        Args:
            terms (list): List of :class:`qibo.hamiltonians.terms.HamiltonianTerm` objects.

        Returns:
            List of :class:`qibo.hamiltonians.terms.TermGroup` objects that contain
            all the given terms.
        """
        # split given terms according to their order
        # order = len(term.target_qubits)
        orders = {}
        for term in terms:
            if len(term) in orders:
                orders[len(term)].append(term)
            else:
                orders[len(term)] = [term]

        groups = []
        # start creating groups with the higher order terms as parents and then
        # append each term of lower order to the first compatible group
        for order in sorted(orders.keys())[::-1]:
            for child in orders[order]:
                flag = True
                for i, group in enumerate(groups):
                    if group.can_append(child):
                        group.append(child)
                        flag = False
                        break
                if flag:
                    groups.append(cls(child))
        return groups

    @property
    def term(self):
        """Returns a single :class:`qibo.hamiltonians.terms.HamiltonianTerm`. after merging all terms in the group."""
        if self._term is None:
            self._term = self.to_term()
        return self._term

    def to_term(self, coefficients={}):
        """Calculates a single :class:`qibo.hamiltonians.terms.HamiltonianTerm` by merging all terms in the group.

        Args:
            coefficients (dict): Optional dictionary that allows passing a different
                coefficient to each term according to its parent Hamiltonian.
                Useful for :class:`qibo.core.adiabatic.AdiabaticHamiltonian` calculations.
        """
        c = coefficients.get(self[0].hamiltonian)
        merged = self[0] * c if c is not None else self[0]
        for term in self[1:]:
            c = coefficients.get(term.hamiltonian)
            merged = merged.merge(term * c if c is not None else term)
        return merged



"""Module defining the Clifford backend."""

import collections
from functools import cache, reduce
from importlib.util import find_spec, module_from_spec
from itertools import product
from typing import Union

import numpy as np

from qibo import gates
from qibo.backends.numpy import NumpyBackend
from qibo.config import raise_error


class CliffordBackend(NumpyBackend):
    """Backend for the simulation of Clifford circuits following
    `Aaronson & Gottesman (2004) <https://arxiv.org/abs/quant-ph/0406196>`_.

    Args:
        :class:`qibo.backends.abstract.Backend`: Backend used for the calculation.
    """

    def __init__(self, engine=None):
        super().__init__()

        if engine == "stim":
            import stim  # pylint: disable=C0415

            engine = "numpy"
            self.platform = "stim"
            self._stim = stim
        else:
            if engine is None:
                from qibo.backends import (  # pylint: disable=C0415
                    _check_backend,
                    _get_engine_name,
                )

                engine = _get_engine_name(_check_backend(engine))

            self.platform = engine

        spec = find_spec("qibo.backends._clifford_operations")
        self.engine = module_from_spec(spec)
        spec.loader.exec_module(self.engine)

        if engine == "numpy":
            pass
        elif engine == "numba":
            from numba import set_num_threads

            set_num_threads(1)

            from qibojit.backends import (  # pylint: disable=C0415
                clifford_operations_cpu,
            )

            for method in dir(clifford_operations_cpu):
                setattr(self.engine, method, getattr(clifford_operations_cpu, method))
        elif engine == "cupy":  # pragma: no cover
            from qibojit.backends import (  # pylint: disable=C0415
                clifford_operations_gpu,
            )

            for method in dir(clifford_operations_gpu):
                setattr(self.engine, method, getattr(clifford_operations_gpu, method))
        else:
            raise_error(
                NotImplementedError,
                f"Backend `{engine}` is not supported for Clifford Simulation.",
            )

        self.np = self.engine.np

        self.name = "clifford"

    def cast(self, x, dtype=None, copy: bool = False):
        """Cast an object as the array type of the current backend.

        Args:
            x: Object to cast to array.
            dtype (optional): data type of the array or tensor. If ``None``, defaults
                to the default data type of the current backend. Defaults to ``None``.
            copy (bool, optional): If ``True`` a copy of the object is created in memory.
                Defaults to ``False``.
        """
        return self.engine.cast(x, dtype=dtype, copy=copy)

    def set_seed(self, seed):
        super().set_seed(seed)
        self.engine.set_seed(seed)

    def calculate_frequencies(self, samples):
        res, counts = self.engine.np.unique(samples, return_counts=True)
        # The next two lines are necessary for the GPU backends
        res = [int(r) if not isinstance(r, str) else r for r in res]
        counts = [int(v) for v in counts]

        return collections.Counter(dict(zip(res, counts)))

    def zero_state(self, nqubits: int, i_phase=False):
        """Construct the zero state :math`\\ket{00...00}`.

        Args:
            nqubits (int): number of qubits.
            i_phase (bool, optional): If ``True``, the symplectic matrix will
                have two phase columns as in Dehaene-De Moor format.
                If ``False``, the symplectic matrix will have one phase column
                as in Aaronson-Gottesman format. Defaults to ``False``.

        Returns:
            ndarray: Symplectic matrix for the zero state.
        """
        identity = self.np.eye(nqubits)
        ncols = 2 * nqubits + 2 if i_phase else 2 * nqubits + 1

        symplectic_matrix = self.np.zeros((2 * nqubits + 1, ncols), dtype=self.np.uint8)
        symplectic_matrix[:nqubits, :nqubits] = self.np.copy(identity)
        symplectic_matrix[nqubits:-1, nqubits : 2 * nqubits] = self.np.copy(identity)
        return symplectic_matrix

    def _clifford_pre_execution_reshape(self, state):
        """Reshape the symplectic matrix to the shape needed by the engine before circuit execution.

        Args:
            state (ndarray): Input state.

        Returns:
            ndarray: Reshaped state.
        """
        return self.engine._clifford_pre_execution_reshape(  # pylint: disable=protected-access
            state
        )

    def _clifford_post_execution_reshape(self, state, nqubits: int):
        """Reshape the symplectic matrix to the shape needed by the engine after circuit execution.

        Args:
            state (ndarray): Input state.
            nqubits (int): Number of qubits.

        Returns:
            ndarray: Reshaped state.
        """
        return self.engine._clifford_post_execution_reshape(  # pylint: disable=protected-access
            state, nqubits
        )

    def apply_gate_clifford(self, gate, symplectic_matrix, nqubits):
        """Apply a gate to a symplectic matrix."""
        if isinstance(gate, gates.Unitary):
            return self.apply_unitary(gate, symplectic_matrix, nqubits)

        operation = getattr(self.engine, gate.__class__.__name__)

        kwargs = {}
        for param_name in ["theta", "phi"]:
            if param_name in gate.init_kwargs:
                kwargs[param_name] = gate.init_kwargs[param_name]

        return operation(symplectic_matrix, *gate.init_args, nqubits, **kwargs)

    def apply_unitary(self, gate, symplectic_matrix, nqubits):
        """Apply a unitary gate to a symplectic matrix following
        `Dehaene & Moor (2003) <https://arxiv.org/abs/quant-ph/0304125>`_."""
        qubit_indices = list(gate.qubits)
        m = len(qubit_indices)
        matrix = gate._parameters[0]
        symplectic_m, phase_h_m = self._compute_symplectic_matrix(matrix, m)
        phase_d_m = self._get_phase_vector_dk(symplectic_m, m)

        symplectic_n = self._embed_clifford(symplectic_m, nqubits, qubit_indices)
        phase_h_n = self._embed_phase_vector(phase_h_m, nqubits, qubit_indices)
        phase_d_n = self._embed_phase_vector(phase_d_m, nqubits, qubit_indices)

        symplectic_gate = [symplectic_n, phase_h_n, phase_d_n]
        for i in range(2 * nqubits):
            symplectic_pauli = [
                symplectic_matrix[i, : 2 * nqubits],
                symplectic_matrix[i, 2 * nqubits],
                symplectic_matrix[i, 2 * nqubits + 1],
            ]
            symplectic_pauli = self._conjugate_pauli(
                symplectic_gate, symplectic_pauli, nqubits
            )

            symplectic_matrix[i, : 2 * nqubits] = symplectic_pauli[0]
            symplectic_matrix[i, 2 * nqubits] = symplectic_pauli[1]
            symplectic_matrix[i, 2 * nqubits + 1] = symplectic_pauli[2]

        return symplectic_matrix

    def _pauli_string_to_matrix(self, pauli_str):
        """Convert Pauli string to matrix (tensor product)."""
        from qibo import matrices  # pylint: disable=C0415

        paulis = {
            pauli: self.engine.cast(getattr(matrices, pauli), dtype=self.dtype)
            for pauli in ("I", "X", "Y", "Z")
        }
        paulis["Y"] = 1j * paulis["Y"]
        pauli_matrices = [paulis.get(p) for p in pauli_str]
        matrix = reduce(self.engine.np.kron, pauli_matrices)
        return matrix

    def _pauli_to_binary(self, pauli_str, nqubits):
        """Convert Pauli string to binary vector of length :math`2*nqubits`."""
        pauli_symplectic = self.np.zeros(2 * nqubits, dtype=self.np.uint8)
        for q, term in enumerate(pauli_str):
            if term in ["X", "Y"]:
                pauli_symplectic[q] = 1
            if term in ["Z", "Y"]:
                pauli_symplectic[q + nqubits] = 1
        return pauli_symplectic

    @staticmethod
    @cache
    def _pauli_generators(m):
        pauli_gens_x, pauli_gens_z = [], []
        for i in range(m):
            p = ["I"] * m
            p = (p, p.copy())
            p[0][i] = "X"
            p[1][i] = "Z"
            pauli_gens_x.append("".join(p[0]))
            pauli_gens_z.append("".join(p[1]))
        return pauli_gens_x + pauli_gens_z

    def _compute_symplectic_matrix(self, unitary, m):
        """Compute the symplectic matrix for Clifford unitary on :math`m` qubits and the phase vector :math`h` of length :math`2m` for Clifford unitary :math`U`.
        :math`h[j] = 0` if :math`U g_j U^\\dagger = i^r p_j` with :math`r=0` or :math`1` else :math`1`.
        """
        pauli_gens = self._pauli_generators(m)

        symplectic = self.np.zeros((2 * m, 2 * m), dtype=self.np.uint8)
        phase_vector = self.np.zeros(2 * m, dtype=self.np.uint8)

        for i, p_str in enumerate(pauli_gens):
            pauli = self._pauli_string_to_matrix(p_str)
            pauli_uconj = unitary @ pauli @ unitary.conj().T

            found = False
            for candidate_str in product("IXYZ", repeat=m):
                candidate_str = "".join(candidate_str)
                candidate_P = self._pauli_string_to_matrix(candidate_str)
                for phase_val, phase_code in zip([1, 1j, -1, -1j], [0, 0, 1, 1]):
                    if self.np.allclose(
                        pauli_uconj, phase_val * candidate_P, atol=1e-10
                    ):
                        phase_vector[i] = phase_code
                        symplectic[i, :] = self._pauli_to_binary(candidate_str, m)
                        found = True
                        break
                if found:
                    break
        return symplectic % 2, phase_vector

    def _get_phase_vector_dk(self, symplectic, m):
        """Compute phase vector :math`d` of length :math`2m` for Clifford unitary :math`U`.
        :math`d[j] = 0` if :math`U g_j U^\\dagger = (-1)^r p_j` with :math`r=0` or :math`1` else :math`1`.
        """
        u_matrix = self.np.zeros((2 * m, 2 * m), dtype=self.np.uint8)
        u_matrix[0:m, m : 2 * m] = self.np.eye(m, dtype=self.np.uint8)
        d = self.np.diag(symplectic @ (u_matrix @ symplectic.T) % 2) % 2
        return d

    def _conjugate_pauli(self, symplectic_gate, symplectic_pauli, nqubits):
        """Compute the conjugate of a Pauli operator under a symplectic transformation."""
        symplectic_matrix, phase_h, phase_d = symplectic_gate
        symplectic_vector, epsilon, delta = symplectic_pauli

        new_symplectic_vector = (symplectic_matrix.T @ symplectic_vector) % 2

        pd_dot_sv = self.np.dot(phase_d, symplectic_vector) % 2
        new_delta = delta ^ pd_dot_sv

        u_matrix = self.np.zeros((2 * nqubits, 2 * nqubits), dtype=self.np.uint8)
        u_matrix[0:nqubits, nqubits : 2 * nqubits] = self.np.eye(
            nqubits, dtype=self.np.uint8
        )

        lows = self.np.tril(
            symplectic_matrix @ (u_matrix @ symplectic_matrix.T)
            ^ self.np.outer(phase_d, phase_d)
        )

        ph_dot_sv = self.np.dot(phase_h, symplectic_vector) % 2
        sv_lows_sv = self.np.dot(symplectic_vector, lows @ symplectic_vector) % 2
        delta_pd_dot_sv = (delta * pd_dot_sv) % 2
        new_epsilon = epsilon ^ ph_dot_sv ^ sv_lows_sv ^ delta_pd_dot_sv
        return new_symplectic_vector, new_epsilon, new_delta

    def _convert_dehaene_to_aaronson(self, dehaene_tableau):
        """
        Convert Dehaene-De Moor tableau with two phase columns to Aaronson-Gottesman format.

        Dehaene-De Moor format:
        - Columns :math`0` to :math`n-1`: X components
        - Columns :math`n` to :math`2n-1`: Z components
        - Column :math`2n`: real phase/sign (:math`0=+1`, :math`1=-1`)
        - Column :math`2n+1`: :math`i`-phase (powers of :math`i`)

        Aaronson-Gottesman format:
        - Columns :math`0` to :math`n-1`: X components
        - Columns :math`n` to :math`2n-1`: Z components
        - Column :math`2n`: real phase (:math`0=+1`, :math`1=-1`)

        Args:
            dehaene_tableau (ndarray): array of shape :math`(2n+1, 2n+2)` in extended Dehaene-De Moor format.

        Returns:
            (ndarray): aaronson_tableau of shape :math`(2n+1, 2n+1)`.
        """
        n_rows, n_cols = dehaene_tableau.shape
        n = (n_cols - 2) // 2

        X_part = dehaene_tableau[:, :n]
        Z_part = dehaene_tableau[:, n : 2 * n]
        real_phases = dehaene_tableau[:, -2]
        i_phases = dehaene_tableau[:, -1]

        y_count = self.np.sum(X_part[:-1] & Z_part[:-1], axis=-1)
        total_i_power = (i_phases[:-1] + y_count) % 4

        final_real_phases = real_phases.copy()
        indices = total_i_power == 2
        final_real_phases[: n_rows - 1][indices] = (
            final_real_phases[: n_rows - 1][indices] + 1
        ) % 2

        aaronson_tableau = np.column_stack([X_part, Z_part, final_real_phases])
        return self.cast(aaronson_tableau, dtype=aaronson_tableau.dtype)

    def _embed_clifford(self, symplectic_m, n, qubit_indices):
        """Embed m-qubit symplectic :math`S_U_m` into n-qubit system at qubit_indices."""
        symplectic_n = self.np.eye(2 * n, dtype=self.np.uint8)

        x_indices = qubit_indices
        z_indices = [q + n for q in qubit_indices]
        full_indices = x_indices + z_indices

        symplectic_n[np.ix_(full_indices, full_indices)] = symplectic_m

        return symplectic_n % 2

    def _embed_phase_vector(self, phase_m, n, qubit_indices):
        """Embed m-qubit phase vector into n-qubit system."""
        phase_n = self.np.zeros(2 * n, dtype=self.np.uint8)
        m = len(qubit_indices)

        qubit_indices = np.array(qubit_indices)
        phase_n[qubit_indices] = phase_m[:m]
        phase_n[qubit_indices + n] = phase_m[m:]

        return phase_n

    def apply_channel(self, channel, state, nqubits):
        probabilities = channel.coefficients + (1 - sum(channel.coefficients),)
        index = self.np.random.choice(
            range(len(probabilities)), size=1, p=probabilities
        )[0]
        index = int(index)
        if index != len(channel.gates):
            gate = channel.gates[index]
            state = gate.apply_clifford(self, state, nqubits)
        return state

    def _execute_circuit_stim(self, circuit, initial_state=None, nshots: int = 1000):
        from qibo.quantum_info.clifford import Clifford  # pylint: disable=C0415

        circuit_stim = self._stim.Circuit()  # pylint: disable=E1101
        for gate in circuit.queue:
            name = gate.__class__.__name__
            name = "S_DAG" if name == "SDG" else name
            circuit_stim.append(name, list(gate.qubits))

        x_destab, z_destab, x_stab, z_stab, x_phases, z_phases = (
            self._stim.Tableau.from_circuit(  # pylint: disable=no-member
                circuit_stim
            ).to_numpy()
        )
        symplectic_matrix = np.block([[x_destab, z_destab], [x_stab, z_stab]])
        symplectic_matrix = np.c_[symplectic_matrix, np.r_[x_phases, z_phases]]

        return Clifford(
            symplectic_matrix,
            measurements=circuit.measurements,
            nshots=nshots,
            _backend=self,
        )

    def execute_circuit(  # pylint: disable=R1710
        self, circuit, initial_state=None, nshots: int = 1000
    ):
        """Execute a Clifford circuits.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Input circuit.
            initial_state (ndarray, optional): The ``symplectic_matrix`` of the initial state.
                If ``None``, defaults to the zero state. Defaults to ``None``.
            nshots (int, optional): Number of shots to perform if ``circuit`` has measurements.
                Defaults to :math:`10^{3}`.

        Returns:
            :class:`qibo.quantum_info.clifford.Clifford`: Object storing to the final results.
        """
        from qibo.quantum_info.clifford import Clifford  # pylint: disable=C0415

        if self.platform == "stim":
            return self._execute_circuit_stim(circuit, initial_state, nshots)

        for gate in circuit.queue:
            if (
                not gate.clifford
                and not gate.__class__.__name__ == "M"
                and not isinstance(gate, gates.PauliNoiseChannel)
            ):
                raise_error(RuntimeError, "Circuit contains non-Clifford gates.")

        if circuit.repeated_execution and nshots != 1:
            return self.execute_circuit_repeated(circuit, nshots, initial_state)

        try:
            nqubits = circuit.nqubits
            i_phase = False
            if any(isinstance(gate, gates.Unitary) for gate in circuit.queue):
                i_phase = True
            state = (
                self.zero_state(nqubits, i_phase)
                if initial_state is None
                else initial_state
            )
            if i_phase is False:
                state = self._clifford_pre_execution_reshape(state)
            for gate in circuit.queue:
                if i_phase:
                    if isinstance(gate, gates.M):
                        if gate.collapse:
                            raise_error(
                                NotImplementedError,
                                "Collapsing measurements with `gates.Unitary` are not implemented in the `CliffordBackend`.",
                            )
                    elif not isinstance(gate, gates.Unitary):
                        gate = gates.Unitary(gate.matrix(backend=self), *gate.qubits)
                gate.apply_clifford(self, state, nqubits)
            if i_phase:
                state = self._convert_dehaene_to_aaronson(state)
            else:
                state = self._clifford_post_execution_reshape(state, nqubits)
            clifford = Clifford(
                state,
                measurements=circuit.measurements,
                nshots=nshots,
                _backend=self,
            )
            return clifford

        except self.oom_error:  # pragma: no cover
            raise_error(
                RuntimeError,
                f"State does not fit in {self.device} memory."
                "Please switch the execution device to a "
                "different one using ``qibo.set_device``.",
            )

    def execute_circuit_repeated(self, circuit, nshots: int = 1000, initial_state=None):
        """Execute a Clifford circuits ``nshots`` times.

        This is used for all the simulations that involve repeated execution.
        For instance when collapsing measurement or noise channels are present.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): input circuit.
            initial_state (ndarray, optional): Symplectic_matrix of the initial state.
                If ``None``, defaults to :meth:`qibo.backends.clifford.CliffordBackend.zero_state`.
                Defaults to ``None``.
            nshots (int, optional): Number of times to repeat the execution.
                Defaults to :math:`1000`.

        Returns:
            :class:`qibo.quantum_info.clifford.Clifford`: Object storing to the final results.
        """
        from qibo.quantum_info.clifford import Clifford  # pylint: disable=C0415

        circuit_copy = circuit.copy()
        samples = []
        for _ in range(nshots):
            res = self.execute_circuit(circuit_copy, initial_state, nshots=1)
            for measurement in circuit_copy.measurements:
                measurement.result.reset()
            samples.append(res.samples())
        samples = self.np.vstack(samples)

        for meas in circuit.measurements:
            meas.result.register_samples(samples[:, meas.target_qubits])

        result = Clifford(
            self.zero_state(circuit.nqubits),
            measurements=circuit.measurements,
            nshots=nshots,
            _backend=self,
        )
        result.symplectic_matrix, result._samples = None, None

        return result

    def sample_shots(
        self,
        state,
        qubits: Union[tuple, list],
        nqubits: int,
        nshots: int,
        collapse: bool = False,
    ):  # pylint: disable=W0221
        """Sample shots by measuring selected ``qubits`` in symplectic matrix of a ``state``.

        Args:
            state (ndarray): symplectic matrix from which to sample shots from.
            qubits: (tuple or list): qubits to measure.
            nqubits (int): total number of qubits of the state.
            nshots (int): number of shots to sample.
            collapse (bool, optional): If ``True`` the input state is going to be
                collapsed with the last shot. Defaults to ``False``.

        Returns:
            (ndarray): Samples shots.
        """
        if isinstance(qubits, list):
            qubits = tuple(qubits)

        if collapse:
            samples = [self.engine.M(state, qubits, nqubits) for _ in range(nshots - 1)]
            samples.append(self.engine.M(state, qubits, nqubits, collapse))
        else:
            samples = [self.engine.M(state, qubits, nqubits) for _ in range(nshots)]
        return self.engine.cast(samples, dtype=int)

    def symplectic_matrix_to_generators(
        self, symplectic_matrix, return_array: bool = False
    ):
        """Extract the stabilizers and destabilizers generators from symplectic matrix.

        Args:
            symplectic_matrix (ndarray): The input symplectic_matrix.
            return_array (bool, optional): If ``True`` returns the generators as ``ndarrays``.
                If ``False``, generators are returned as strings. Defaults to ``False``.

        Returns:
            (list, list): Extracted generators and their corresponding phases, respectively.
        """
        bits_to_gate = {"00": "I", "01": "X", "10": "Z", "11": "Y"}

        nqubits = int((symplectic_matrix.shape[1] - 1) / 2)
        phases = (-1) ** symplectic_matrix[:-1, -1].astype(np.int16)
        tmp = 1 * symplectic_matrix[:-1, :-1]
        X, Z = tmp[:, :nqubits], tmp[:, nqubits:]
        generators = []
        for x, z in zip(X, Z):
            paulis = [bits_to_gate[f"{zz}{xx}"] for xx, zz in zip(x, z)]
            if return_array:
                from qibo import matrices  # pylint: disable=C0415

                paulis = [self.cast(getattr(matrices, p)) for p in paulis]
                matrix = reduce(self.np.kron, paulis)
                generators.append(matrix)
            else:
                generators.append("".join(paulis))

        if return_array:
            generators = self.cast(generators)

        return generators, phases



"""Module defining Hamiltonian classes."""

import operator
from functools import cache, cached_property, reduce
from itertools import chain
from math import prod
from typing import Optional

import numpy as np
import sympy

from qibo.backends import Backend, _check_backend
from qibo.config import log, raise_error
from qibo.hamiltonians.abstract import AbstractHamiltonian
from qibo.hamiltonians.terms import SymbolicTerm
from qibo.symbols import Symbol, Z


class Hamiltonian(AbstractHamiltonian):
    """Hamiltonian based on a dense or sparse matrix representation.

    Args:
        nqubits (int): number of quantum bits.
        matrix (ndarray): Matrix representation of the Hamiltonian in the
            computational basis as an array of shape :math:`2^{n} \\times 2^{n}`.
            Sparse matrices based on ``scipy.sparse`` for ``numpy`` / ``qibojit`` backends
            (or on ``tf.sparse`` for the ``tensorflow`` backend) are also supported.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.
    """

    def __init__(self, nqubits, matrix, backend=None):
        from qibo.backends import _check_backend

        self.backend = _check_backend(backend)

        if not (
            isinstance(matrix, self.backend.tensor_types)
            or self.backend.is_sparse(matrix)
        ):
            raise_error(
                TypeError,
                f"Matrix of invalid type {type(matrix)} given during Hamiltonian initialization",
            )
        matrix = self.backend.cast(matrix)

        super().__init__()
        self.nqubits = nqubits
        self.matrix = matrix
        self._eigenvalues = None
        self._eigenvectors = None
        self._exp = {"a": None, "result": None}

    @property
    def matrix(self):
        """Returns the full matrix representation.

        For :math:`n` qubits, can be a dense :math:`2^{n} \\times 2^{n}` array or a sparse
        matrix, depending on how the Hamiltonian was created.
        """
        return self._matrix

    @matrix.setter
    def matrix(self, mat):
        shape = tuple(mat.shape)
        if shape != 2 * (2**self.nqubits,):
            raise_error(
                ValueError,
                f"The Hamiltonian is defined for {self.nqubits} qubits "
                + f"while the given matrix has shape {shape}.",
            )
        self._matrix = mat

    def eigenvalues(self, k=6):
        if self._eigenvalues is None:
            self._eigenvalues = self.backend.calculate_eigenvalues(self.matrix, k)
        return self._eigenvalues

    def eigenvectors(self, k=6):
        if self._eigenvectors is None:
            self._eigenvalues, self._eigenvectors = self.backend.calculate_eigenvectors(
                self.matrix, k
            )
        return self._eigenvectors

    def exp(self, a):
        from qibo.quantum_info.linalg_operations import (  # pylint: disable=C0415
            matrix_exponentiation,
        )

        if self._exp.get("a") != a:
            self._exp["a"] = a
            self._exp["result"] = matrix_exponentiation(
                self.matrix,
                -1j * a,
                self._eigenvectors,
                self._eigenvalues,
                self.backend,
            )
        return self._exp.get("result")

    def expectation(self, state, normalize=False):
        if isinstance(state, self.backend.tensor_types):
            state = self.backend.cast(state)
            shape = tuple(state.shape)
            if len(shape) == 1:  # state vector
                return self.backend.calculate_expectation_state(self, state, normalize)

            if len(shape) == 2:  # density matrix
                return self.backend.calculate_expectation_density_matrix(
                    self, state, normalize
                )

            raise_error(
                ValueError,
                "Cannot calculate Hamiltonian expectation value "
                + f"for state of shape {shape}",
            )

        raise_error(
            TypeError,
            "Cannot calculate Hamiltonian expectation "
            + f"value for state of type {type(state)}",
        )

    def expectation_from_samples(self, freq, qubit_map=None):
        obs = self.matrix
        diag = self.backend.np.diagonal(obs)
        if self.backend.np.count_nonzero(obs - self.backend.np.diag(diag)) != 0:
            raise_error(
                NotImplementedError,
                "Observable is not diagonal. Expectation of non diagonal observables starting from samples is currently supported for `qibo.hamiltonians.hamiltonians.SymbolicHamiltonian` only.",
            )
        diag = self.backend.np.reshape(diag, self.nqubits * (2,))
        if qubit_map is None:
            qubit_map = range(self.nqubits)
        diag = self.backend.np.transpose(diag, qubit_map).ravel()
        # select only the elements with non-zero counts
        diag = diag[[int(state, 2) for state in freq.keys()]]
        counts = self.backend.cast(list(freq.values()), dtype=diag.dtype) / sum(
            freq.values()
        )
        return self.backend.np.real(self.backend.np.sum(diag * counts))

    def eye(self, dim: Optional[int] = None):
        """Generate Identity matrix with dimension ``dim``"""
        if dim is None:
            dim = int(self.matrix.shape[0])
        return self.backend.cast(self.backend.matrices.I(dim), dtype=self.matrix.dtype)

    def energy_fluctuation(self, state):
        """
        Evaluate energy fluctuation:

        .. math::
            \\Xi_{k}(\\mu) = \\sqrt{\\bra{\\mu} \\, H^{2} \\, \\ket{\\mu}
                - \\bra{\\mu} \\, H \\, \\ket{\\mu}^2} \\, .

        for a given state :math:`\\ket{\\mu}`.

        Args:
            state (ndarray): quantum state to be used to compute the energy fluctuation.

        Returns:
            float: Energy fluctuation value.
        """
        state = self.backend.cast(state)
        energy = self.expectation(state)
        h = self.matrix
        h2 = Hamiltonian(nqubits=self.nqubits, matrix=h @ h, backend=self.backend)
        average_h2 = self.backend.calculate_expectation_state(h2, state, normalize=True)
        return self.backend.np.sqrt(self.backend.np.abs(average_h2 - energy**2))

    def __add__(self, other):
        if isinstance(other, self.__class__):
            if self.nqubits != other.nqubits:
                raise_error(
                    RuntimeError,
                    "Only hamiltonians with the same number of qubits can be added.",
                )
            new_matrix = self.matrix + other.matrix
        elif isinstance(other, self.backend.numeric_types) or isinstance(
            other, self.backend.tensor_types
        ):
            new_matrix = self.matrix + other * self.eye()
        else:
            raise_error(
                NotImplementedError,
                f"Hamiltonian addition to {type(other)} not implemented.",
            )
        return self.__class__(
            self.nqubits, new_matrix, backend=self.backend  # pylint: disable=E0606
        )

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            if self.nqubits != other.nqubits:
                raise_error(
                    RuntimeError,
                    "Only hamiltonians with the same number of qubits can be subtracted.",
                )
            new_matrix = self.matrix - other.matrix
        elif isinstance(other, self.backend.numeric_types):
            new_matrix = self.matrix - other * self.eye()
        else:
            raise_error(
                NotImplementedError,
                f"Hamiltonian subtraction to {type(other)} not implemented.",
            )
        return self.__class__(
            self.nqubits, new_matrix, backend=self.backend  # pylint: disable=E0606
        )

    def __rsub__(self, other):
        if isinstance(other, self.__class__):  # pragma: no cover
            # impractical case because it will be handled by `__sub__`
            if self.nqubits != other.nqubits:
                raise_error(
                    RuntimeError,
                    "Only hamiltonians with the same number of qubits can be added.",
                )
            new_matrix = other.matrix - self.matrix
        elif isinstance(other, self.backend.numeric_types):
            new_matrix = other * self.eye() - self.matrix
        else:
            raise_error(
                NotImplementedError,
                f"Hamiltonian subtraction to {type(other)} not implemented.",
            )
        return self.__class__(
            self.nqubits, new_matrix, backend=self.backend  # pylint: disable=E0606
        )

    def __mul__(self, other):
        if isinstance(other, self.backend.tensor_types):
            other = complex(other)
        elif not isinstance(other, self.backend.numeric_types):
            raise_error(
                NotImplementedError,
                f"Hamiltonian multiplication to {type(other)} not implemented.",
            )
        new_matrix = self.matrix * other
        r = self.__class__(self.nqubits, new_matrix, backend=self.backend)
        other = self.backend.cast(other)
        if self._eigenvalues is not None:
            if self.backend.np.real(other) >= 0:  # TODO: check for side effects K.qnp
                r._eigenvalues = other * self._eigenvalues
            elif not self.backend.is_sparse(self.matrix):
                axis = (0,) if (self.backend.platform == "pytorch") else 0
                r._eigenvalues = other * self.backend.np.flip(self._eigenvalues, axis)
        if self._eigenvectors is not None:
            if self.backend.np.real(other) > 0:  # TODO: see above
                r._eigenvectors = self._eigenvectors
            elif other == 0:
                r._eigenvectors = self.eye(int(self._eigenvectors.shape[0]))
        return r

    def __matmul__(self, other):
        if isinstance(other, self.__class__):
            matrix = self.matrix @ other.matrix
            return self.__class__(self.nqubits, matrix, backend=self.backend)

        if isinstance(other, self.backend.tensor_types):
            return self.matrix @ other

        raise_error(
            NotImplementedError,
            f"Hamiltonian matmul to {type(other)} not implemented.",
        )


def _calculate_nqubits_from_form(form):
    """Calculate number of qubits in the system described by the given
    Hamiltonian formula
    """
    nqubits = 0
    for symbol in form.free_symbols:
        if isinstance(symbol, Symbol):
            q = symbol.target_qubit
        else:
            raise_error(
                RuntimeError,
                f"Symbol {symbol} is not a ``qibo.symbols.Symbol``, you can define a custom symbol for {symbol} by subclassing ``qibo.symbols.Symbol``.",
            )
        if q > nqubits:  # pylint: disable=E0606
            nqubits = q
    return nqubits + 1


class SymbolicHamiltonian(AbstractHamiltonian):
    """Hamiltonian based on a symbolic representation.

    Calculations using symbolic Hamiltonians are either done directly using
    the given ``sympy`` expression as it is (``form``) or by parsing the
    corresponding ``terms`` (which are :class:`qibo.core.terms.SymbolicTerm`
    objects). The latter approach is more computationally costly as it uses
    a ``sympy.expand`` call on the given form before parsing the terms.
    For this reason the ``terms`` are calculated only when needed, for example
    during Trotterization.
    The dense matrix of the symbolic Hamiltonian can be calculated directly
    from ``form`` without requiring ``terms`` calculation (see
    :meth:`qibo.core.hamiltonians.SymbolicHamiltonian.calculate_dense` for details).

    Args:
        form (sympy.Expr): Hamiltonian form as a ``sympy.Expr``. Ideally the
            Hamiltonian should be written using Qibo symbols.
            See :ref:`How to define custom Hamiltonians using symbols? <symbolicham-example>`
            example for more details.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.
    """

    def __init__(
        self,
        form: sympy.Expr,
        nqubits: Optional[int] = None,
        backend: Optional[Backend] = None,
    ):
        super().__init__()
        if not isinstance(form, sympy.Expr):
            raise_error(
                TypeError,
                f"The ``form`` of a ``SymbolicHamiltonian`` has to be a ``sympy.Expr``, but a ``{type(form)}`` was passed.",
            )
        self._form = form
        self.constant = 0  # used only when we perform calculations using ``_terms``

        self.backend = _check_backend(backend)

        self.nqubits = (
            _calculate_nqubits_from_form(form) if nqubits is None else nqubits
        )

    @cached_property
    def dense(self) -> "MatrixHamiltonian":  # type: ignore
        """Creates the equivalent Hamiltonian matrix."""
        return self.calculate_dense()

    @property
    def form(self):
        return self._form

    @form.setter
    def form(self, form):
        # Check that given form is a ``sympy`` expression
        if not isinstance(form, sympy.Expr):
            raise_error(
                TypeError,
                f"Symbolic Hamiltonian should be a ``sympy`` expression but is {type(form)}.",
            )
        self._form = form
        self.nqubits = _calculate_nqubits_from_form(form)

    @cached_property
    def terms(self):
        """List of terms of which the Hamiltonian is a sum of.

        Terms will be objects of type :class:`qibo.core.terms.HamiltonianTerm`.
        """
        # Calculate terms based on ``self.form``

        self.constant = 0.0

        form = sympy.expand(self.form)
        terms = []
        for factors, coeff in form.as_coefficients_dict().items():
            term = SymbolicTerm(coeff, factors, backend=self.backend)
            if term.target_qubits:
                # Check for any terms with the same factors and add their coefficients together
                found = False
                for i, _term in enumerate(terms):
                    if set(_term.factors) == set(term.factors):
                        found = True
                        terms[i].coefficient += term.coefficient
                        break
                if not found:
                    terms.append(term)
            else:
                self.constant += term.coefficient
        return terms

    @cached_property
    def diagonal_terms(self) -> list[list[SymbolicTerm]]:
        """List of terms that can be diagonalized simultaneously, i.e. that
        commute with each other. In detail each element of the list is a sublist
        of commuting ``SymbolicTerm``s.
        """
        diagonal_terms = []
        terms = self.terms
        # loop until there are no more terms
        while len(terms) > 0:
            # take the first (remaining) term
            t0 = terms[0]
            diagonal_term = [t0]
            removable_indices = {
                0,
            }
            # look for all the following terms that commute with the
            # first one and among themselves
            for i, t1 in enumerate(terms[1:], 1):
                # commutes with all the other terms -> append it
                if all(term.commute(t1) for term in diagonal_term):
                    diagonal_term.append(t1)
                    removable_indices.add(i)
            # append the new sublist of commuting terms found
            diagonal_terms.append(diagonal_term)
            # remove them from the original terms
            terms = [term for i, term in enumerate(terms) if i not in removable_indices]
        return diagonal_terms

    @property
    def matrix(self):
        """Returns the full matrix representation.

        Consisting of :math:`2^{n} \\times 2^{n}`` elements.
        """
        return self.dense.matrix

    def eigenvalues(self, k=6):
        return self.dense.eigenvalues(k)

    def eigenvectors(self, k=6):
        return self.dense.eigenvectors(k)

    def ground_state(self):
        return self.eigenvectors()[:, 0]

    def exp(self, a):
        return self.dense.exp(a)

    @cache
    def _get_symbol_matrix(self, term):
        """Calculates numerical matrix corresponding to symbolic expression.

        This is partly equivalent to sympy's ``.subs``, which does not work
        in our case as it does not allow us to substitute ``sympy.Symbol``
        with numpy arrays and there are different complication when switching
        to ``sympy.MatrixSymbol``. Here we calculate the full numerical matrix
        given the symbolic expression using recursion.
        Helper method for ``_calculate_dense_from_form``.

        Args:
            term (sympy.Expr): Symbolic expression containing local operators.

        Returns:
            ndarray: matrix corresponding to the given expression as an array
            of shape ``(2 ** self.nqubits, 2 ** self.nqubits)``.
        """
        if isinstance(term, sympy.Add):
            # symbolic op for addition
            result = sum(
                self._get_symbol_matrix(subterm) for subterm in term.as_ordered_terms()
            )

        elif isinstance(term, sympy.Mul):
            # symbolic op for multiplication
            # note that we need to use matrix multiplication even though
            # we use scalar symbols for convenience
            factors = term.as_ordered_factors()
            result = reduce(
                self.backend.np.matmul,
                (self._get_symbol_matrix(subterm) for subterm in factors),
            )

        elif isinstance(term, sympy.Pow):
            # symbolic op for power
            base, exponent = term.as_base_exp()
            matrix = self._get_symbol_matrix(base)
            matrix_power = (
                np.linalg.matrix_power
                if self.backend.name == "tensorflow"
                else self.backend.np.linalg.matrix_power
            )
            result = matrix_power(matrix, int(exponent))

        elif isinstance(term, Symbol):
            # if we have a Qibo symbol the matrix construction is
            # implemented in :meth:`qibo.core.terms.SymbolicTerm.full_matrix`.
            # I have to force the symbol's backend
            term.backend = self.backend
            result = term.full_matrix(self.nqubits)

        elif term.is_number:
            # if the term is number we should return in the form of identity
            # matrix because in expressions like `1 + Z`, `1` is not correspond
            # to the float 1 but the identity operator (matrix)
            result = complex(term) * self.backend.matrices.I(2**self.nqubits)

        else:
            raise_error(
                TypeError,
                f"Cannot calculate matrix for symbolic term of type {type(term)}.",
            )

        return result  # pylint: disable=E0606

    def _calculate_dense_from_form(self) -> Hamiltonian:
        """Calculates equivalent Hamiltonian using symbolic form.

        Useful when the term representation is not available.
        """
        matrix = self._get_symbol_matrix(self.form)
        return Hamiltonian(self.nqubits, matrix, backend=self.backend)

    def calculate_dense(self) -> Hamiltonian:
        log.warning(
            "Calculating the dense form of a symbolic Hamiltonian. "
            "This operation is memory inefficient."
        )
        # calculate dense matrix directly using the form to avoid the
        # costly ``sympy.expand`` call
        return self._calculate_dense_from_form()

    def expectation(self, state, normalize=False):
        return Hamiltonian.expectation(self, state, normalize)

    def expectation_from_circuit(self, circuit: "Circuit", nshots: int = 1000) -> float:  # type: ignore
        """
        Calculate the expectation value from a circuit.
        This even works for observables not completely diagonal in the computational
        basis, but only diagonal at a term level in a defined basis. Namely, for
        an observable of the form :math:``H = \\sum_i H_i``, where each :math:``H_i``
        consists in a `n`-qubits pauli operator :math:`P_0 \\otimes P_1 \\otimes \\cdots \\otimes P_n`,
        the expectation value is computed by rotating the input circuit in the suitable
        basis for each term :math:``H_i`` thus extracting the `term-wise` expectations
        that are then summed to build the global expectation value.
        Each term of the observable is treated separately, by measuring in the correct
        basis and re-executing the circuit.

        Args:
            circuit (Circuit): input circuit.
            nshots (int): number of shots, defaults to 1000.

        Returns:
            (float): the calculated expectation value.
        """
        from qibo import gates

        rotated_circuits = []
        Z_observables = []
        qubit_maps = []
        # loop over the terms that can be diagonalized simultaneously
        for terms in self.diagonal_terms:
            # for each term that can be diagonalized simultaneously
            # extract the coefficient, Z observable and qubit map
            # the basis rotation, instead, will be the same
            tmp_obs = []
            measurements = {}
            for term in terms:
                # Only care about non-I terms
                non_identity_factors = []
                # prepare the measurement basis and append it to the circuit
                for factor in term.factors:
                    if factor.name[0] != "I":
                        non_identity_factors.append(factor)
                        q = factor.target_qubit
                        if q not in measurements:
                            measurements[q] = gates.M(q, basis=factor.gate.__class__)

                # build diagonal observable
                # including the coefficient
                symb_obs = prod(
                    Z(factor.target_qubit) for factor in non_identity_factors
                )
                symb_obs = SymbolicHamiltonian(
                    term.coefficient * symb_obs,
                    nqubits=circuit.nqubits,
                    backend=self.backend,
                )
                tmp_obs.append(symb_obs)

            # Get the qubits we want to measure for each term
            qubit_maps.append(sorted(measurements.keys()))
            Z_observables.append(tmp_obs)

            circ_copy = circuit.copy(True)
            circ_copy.add(list(measurements.values()))
            rotated_circuits.append(circ_copy)

        # execute the circuits
        results = self.backend.execute_circuits(rotated_circuits, nshots=nshots)

        # construct the expectation value for each diagonal term
        # and sum all together
        expval = 0.0
        for res, obs, qmap in zip(results, Z_observables, qubit_maps):
            freq = res.frequencies()
            expval += sum(o.expectation_from_samples(freq, qmap) for o in obs)
        return self.constant + expval

    def expectation_from_samples(self, freq: dict, qubit_map: list = None) -> float:
        """
        Calculate the expectation value from the samples.
        The observable has to be diagonal in the computational basis.

        Args:
            freq (dict): input frequencies of the samples.
            qubit_map (list): qubit map.

        Returns:
            (float): the calculated expectation value.
        """
        for term in self.terms:
            # pylint: disable=E1101
            for factor in term.factors:
                if not isinstance(factor, Z):
                    raise_error(
                        NotImplementedError, "Observable is not a Pauli-Z string."
                    )

        if qubit_map is None:
            qubit_map = list(range(self.nqubits))

        keys = list(freq.keys())
        counts = list(freq.values())
        counts = self.backend.cast(counts, dtype=self.backend.np.float64) / sum(counts)
        expvals = []
        for term in self.terms:
            qubits = {
                factor.target_qubit for factor in term.factors if factor.name[0] != "I"
            }
            expvals.extend(
                [
                    term.coefficient.real
                    * (-1) ** [state[qubit_map.index(q)] for q in qubits].count("1")
                    for state in keys
                ]
            )
        expvals = self.backend.cast(expvals, dtype=counts.dtype).reshape(
            len(self.terms), len(freq)
        )
        return self.backend.np.sum(expvals @ counts) + self.constant.real

    def _compose(self, other, operator):
        form = self._form

        if isinstance(other, self.__class__):
            if self.nqubits != other.nqubits:
                raise_error(
                    RuntimeError,
                    "Only Hamiltonians with the same number of qubits can be composed.",
                )

            if other._form is not None:
                form = operator(form, other._form) if form is not None else other._form

        elif isinstance(other, (self.backend.numeric_types, self.backend.tensor_types)):
            form = (
                operator(form, complex(other)) if form is not None else complex(other)
            )
        else:
            raise_error(
                NotImplementedError,
                f"SymbolicHamiltonian composition to {type(other)} not implemented.",
            )

        return self.__class__(form=form, nqubits=self.nqubits, backend=self.backend)

    def __add__(self, other):
        return self._compose(other, operator.add)

    def __sub__(self, other):
        return self._compose(other, operator.sub)

    def __rsub__(self, other):
        return self._compose(other, lambda x, y: y - x)

    def __mul__(self, other):
        return self._compose(other, lambda x, y: y * x)

    def apply_gates(self, state, density_matrix=False):
        """Applies gates corresponding to the Hamiltonian terms.

        Gates are applied to the given state.

        Helper method for :meth:`qibo.hamiltonians.SymbolicHamiltonian.__matmul__`.
        """
        total = 0
        for term in self.terms:
            total += term(
                self.backend,
                self.backend.cast(state, copy=True),
                self.nqubits,
                density_matrix=density_matrix,
            )
        if self.constant:  # pragma: no cover
            total += self.constant * state
        return total

    def __matmul__(self, other):
        """Matrix multiplication with other Hamiltonians or state vectors."""
        if isinstance(other, self.__class__):
            return other * self

        if isinstance(other, self.backend.tensor_types):
            rank = len(tuple(other.shape))
            if rank not in (1, 2):
                raise_error(
                    NotImplementedError,
                    f"Cannot multiply Hamiltonian with rank-{rank} tensor.",
                )
            state_qubits = int(np.log2(int(other.shape[0])))
            if state_qubits != self.nqubits:
                raise_error(
                    ValueError,
                    f"Cannot multiply Hamiltonian on {self.nqubits} qubits to "
                    + f"state of {state_qubits} qubits.",
                )
            if rank == 1:  # state vector
                return self.apply_gates(other)

            return self.apply_gates(other, density_matrix=True)

        raise_error(
            NotImplementedError,
            f"Hamiltonian matmul to {type(other)} not implemented.",
        )

    def circuit(self, dt, accelerators=None):
        """Circuit that implements a Trotter step of this Hamiltonian.

        Args:
            dt (float): Time step used for Trotterization.
            accelerators (dict, optional): Dictionary with accelerators for distributed circuits.
                Defaults to ``None``.
        """
        from qibo import Circuit  # pylint: disable=import-outside-toplevel
        from qibo.hamiltonians.terms import (  # pylint: disable=import-outside-toplevel
            TermGroup,
        )

        groups = TermGroup.from_terms(self.terms)
        circuit = Circuit(self.nqubits, accelerators=accelerators)
        circuit.add(
            group.term.expgate(dt / 2.0) for group in chain(groups, groups[::-1])
        )

        return circuit



"""Distances, metrics, and measures for quantum states and channels."""

from typing import Optional, Union

import numpy as np
from scipy import sparse

from qibo.backends import _check_backend
from qibo.config import PRECISION_TOL, raise_error


def purity(state, backend=None):
    """Purity of a quantum state :math:`\\rho`.

    This is given by

    .. math::
        \\text{purity}(\\rho) = \\text{tr}(\\rho^{2}) \\, .

    Args:
        state (ndarray): statevector or density matrix.
    Returns:
        float: Purity of quantum ``state`` :math:`\\rho`.
    """
    backend = _check_backend(backend)
    if (
        (len(state.shape) >= 3)
        or (len(state) == 0)
        or (len(state.shape) == 2 and state.shape[0] != state.shape[1])
    ):
        raise_error(
            TypeError,
            f"state must have dims either (k,) or (k,k), but have dims {state.shape}.",
        )

    if len(state.shape) == 1:
        pur = backend.np.real(backend.calculate_vector_norm(state)) ** 2
    else:
        pur = backend.np.real(backend.np.trace(backend.np.matmul(state, state)))
    return float(pur)


def impurity(state, backend=None):
    """Impurity of quantum state :math:`\\rho`.

    This is given by :math:`1 - \\text{purity}(\\rho)`, where :math:`\\text{purity}`
    is defined in :func:`qibo.quantum_info.purity`.

    Args:
        state (ndarray): statevector or density matrix.

    Returns:
        float: impurity of ``state`` :math:`\\rho`.
    """
    return 1 - purity(state, backend=backend)


def trace_distance(state, target, backend=None):
    """Trace distance between two quantum states, :math:`\\rho` and
    :math:`\\sigma`:

    .. math::
        T(\\rho, \\sigma) = \\frac{1}{2} \\, \\|\\rho - \\sigma\\|_{1} = \\frac{1}{2} \\,
            \\text{tr}\\left[ \\sqrt{(\\rho - \\sigma)^{\\dagger}(\\rho - \\sigma)}
            \\right] \\, ,

    where :math:`\\|\\cdot\\|_{1}` is the Schatten 1-norm.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Trace distance between ``state`` :math:`\\rho` and ``target`` :math:`\\sigma`.
    """
    backend = _check_backend(backend)

    if state.shape != target.shape:
        raise_error(
            TypeError,
            f"State has dims {state.shape} while target has dims {target.shape}.",
        )

    if (len(state.shape) >= 3) or (len(state) == 0):
        raise_error(
            TypeError,
            "Both objects must have dims either (k,) or (k,l), "
            + f"but have dims {state.shape} and {target.shape}",
        )

    if len(state.shape) == 1:
        state = backend.np.outer(backend.np.conj(state), state)
        target = backend.np.outer(backend.np.conj(target), target)

    distance = state - target
    distance = backend.np.conj(distance.T) @ distance
    distance = backend.calculate_matrix_sqrt(distance)

    return backend.np.trace(distance) / 2


def hilbert_schmidt_inner_product(operator_A, operator_B, backend=None):
    """Calculate the Hilbert-Schmidt inner product between two operators.

    Given two operators :math:`A, \\, B \\in \\mathcal{H}`, the Hilbert-Schmidt
    inner product between the two is given by

    .. math::
        \\braket{A, \\, B}_{\\text{HS}} = \\text{tr}\\left(A^{\\dagger} \\, B\\right) \\, .

    Args:
        operator_A (ndarray): operator :math:`A`.
        operator_B (ndarray): operator :math:`B`.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Hilbert-Schmidt inner product :math:`\\braket{A, \\, B}_{\\text{HS}}`.
    """
    backend = _check_backend(backend)

    inner_product = backend.np.trace(backend.np.conj(operator_A.T) @ operator_B)

    return backend.np.real(inner_product)


def hilbert_schmidt_distance(state, target, backend=None):
    """Calculate the Hilbert-Schmidt distance between two quantum states:

    .. math::
        \\braket{\\rho - \\sigma, \\, \\rho - \\sigma}_{\\text{HS}} =
            \\text{tr}\\left((\\rho - \\sigma)^{2}\\right) \\, ,

    where :math:`\\braket{\\cdot, \\, \\cdot}_{\\text{HS}}` is the
    :func:`qibo.quantum_info.hilbert_schmidt_inner_product`.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Hilbert-Schmidt distance between ``state`` :math:`\\rho`
        and ``target`` :math:`\\sigma`.

    References:
        1. P. J. Coles, M. Cerezo, and L. Cincio, *Strong bound between trace distance
        and Hilbert-Schmidt distance for low-rank states*, `Phys. Rev. A 100, 022103
        <https://doi.org/10.1103/PhysRevA.100.022103>`_ (2019).
    """
    backend = _check_backend(backend)

    if state.shape != target.shape:
        raise_error(
            TypeError,
            f"State has dims {state.shape} while target has dims {target.shape}.",
        )

    if (len(state.shape) >= 3) or (len(state) == 0):
        raise_error(
            TypeError,
            "Both objects must have dims either (k,) or (k,l), "
            + f"but have dims {state.shape} and {target.shape}",
        )

    if len(state.shape) == 1:
        state = backend.np.outer(backend.np.conj(state), state)
        target = backend.np.outer(backend.np.conj(target), target)

    difference = state - target

    return hilbert_schmidt_inner_product(difference, difference, backend=backend)


def fidelity(state, target, backend=None):
    """Fidelity :math:`F(\\rho, \\sigma)` between ``state`` :math:`\\rho` and
    ``target`` state :math:`\\sigma`. In general,

    .. math::
        F(\\rho, \\sigma) = \\text{tr}^{2}\\left( \\sqrt{\\sqrt{\\sigma} \\,
        \\rho^{\\dagger} \\, \\sqrt{\\sigma}} \\right) \\, .

    However, when at least one of the states is pure, then

    .. math::
        F(\\rho, \\sigma) = \\text{tr}(\\rho \\, \\sigma)

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Fidelity between ``state`` :math:`\\rho` and ``target`` :math:`\\sigma`.
    """
    backend = _check_backend(backend)
    state = backend.cast(state, dtype=state.dtype)
    target = backend.cast(target, dtype=target.dtype)
    if state.shape != target.shape:
        raise_error(
            TypeError,
            f"State has dims {state.shape} while target has dims {target.shape}.",
        )

    if len(state.shape) >= 3 or len(state.shape) == 0:
        raise_error(
            TypeError,
            "Both objects must have dims either (k,) or (k,l), "
            + f"but have dims {state.shape} and {target.shape}",
        )

    # check purity if both states are density matrices
    if len(state.shape) == 2 and len(target.shape) == 2:
        purity_state = purity(state, backend=backend)
        purity_target = purity(target, backend=backend)

        # if both states are mixed, default to full fidelity calculation
        if (
            abs(purity_state - 1) > PRECISION_TOL
            and abs(purity_target - 1) > PRECISION_TOL
        ):
            fid = backend.calculate_matrix_sqrt(state)
            fid = fid @ backend.np.conj(target.T) @ fid
            fid = backend.calculate_matrix_sqrt(fid)
            fid = backend.np.real(backend.np.trace(fid))

            return fid**2

    # if any of the states is pure, perform lighter calculation
    fid = (
        backend.np.abs(backend.np.matmul(backend.np.conj(state), target)) ** 2
        if len(state.shape) == 1
        else backend.np.real(backend.np.trace(backend.np.matmul(state, target)))
    )

    return fid


def infidelity(state, target, backend=None):
    """Infidelity between ``state`` :math:`\\rho` and ``target`` state
    :math:`\\sigma`, which is given by

    .. math::
        1 - F(\\rho, \\, \\sigma) \\, ,

    where :math:`F(\\rho, \\, \\sigma)` is the :func:`qibo.quantum_info.fidelity`
    between ``state`` and ``target``.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Infidelity between ``state`` :math:`\\rho` and ``target`` :math:`\\sigma`.
    """
    return 1 - fidelity(state, target, backend=backend)


def a_fidelity(state, target, backend=None):
    """Return the :math:`A`-fidelity between two quantum states.

    For a quantum ``state`` :math:`\\rho` and a ``target`` quantum state :math:`\\sigma`,
    the :math:`A`-fidelity is defined as:

    .. math::
        F_{\\text{A}}(\\rho, \\, \\sigma) = \\text{tr}^{2}(\\sqrt{\\rho} \\, \\sqrt{\\sigma}) \\, .

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: :math:`A`-fidelity between :math:`\\rho` and :math:`\\sigma`.

    References:
        1. Y.-C. Liang, Y.-H. Yeh, P. E. M. F. Mendonça, R. Y. Teh, M. D. Reid, and P. D. Drummond,
        *Quantum fidelity measures for mixed states*, `Rep. Prog. Phys. 82 076001 (2019)
        <https://iopscience.iop.org/article/10.1088/1361-6633/ab1ca4>`_.

        2. Z. Ma, F.-L. Zhang, and J.-L. Chen. *Geometric interpretation for the A ﬁdelity and its
        relation with the Bures ﬁdelity*, `Phys. Rev. A, 78(6):064305, (2008)
        <https://doi.org/10.1103/PhysRevA.78.064305>`_.
    """

    backend = _check_backend(backend)

    state = backend.cast(state, dtype=state.dtype)
    target = backend.cast(target, dtype=target.dtype)

    purity_state = purity(state, backend=backend)
    purity_target = purity(target, backend=backend)

    test_state = bool(backend.np.abs(purity_state - 1) <= PRECISION_TOL)
    test_target = bool(backend.np.abs(purity_target - 1) <= PRECISION_TOL)

    if test_state and test_target:
        return fidelity(state, target, backend=backend) ** 2

    state_sqrt = backend.calculate_matrix_sqrt(state) if not test_state else state
    target_sqrt = backend.calculate_matrix_sqrt(target) if not test_target else target

    if test_state and not test_target:
        trace = (
            backend.np.conj(state) @ target_sqrt @ state
            if len(state.shape) == 1
            else backend.np.trace(state @ target_sqrt)
        )
        return backend.np.real(trace) ** 2

    if not test_state and test_target:
        trace = (
            backend.np.conj(target) @ state_sqrt @ target
            if len(target.shape) == 1
            else backend.np.trace(state_sqrt @ target)
        )
        return backend.np.real(trace) ** 2

    return backend.np.real(backend.np.trace(state_sqrt @ target_sqrt)) ** 2


def n_fidelity(state, target, backend=None):
    """Return the :math:`N`-fidelity between two quantum states.

    For a quantum ``state`` :math:`\\rho` and a ``target`` quantum state :math:`\\sigma`,
    the :math:`N`-fidelity is defined as:

    .. math::
        F_{\\text{N}}(\\rho, \\, \\sigma) = \\text{tr}(\\rho \\, \\sigma) +
            \\sqrt{1 - \\text{tr}(\\rho^{2})} \\, \\sqrt{1 - \\text{tr}(\\rho^{2})} \\, ,

    where :math:`\\text{tr}(\\varrho^{2})` is the :class:`qibo.quantum_info.purity` of
    a quantum state :math:`\\varrho`.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: :math:`N`-fidelity between :math:`\\rho` and :math:`\\sigma`.

    References:
        1. Y.-C. Liang, Y.-H. Yeh, P. E. M. F. Mendonça, R. Y. Teh, M. D. Reid, and P. D. Drummond,
        *Quantum fidelity measures for mixed states*, `Rep. Prog. Phys. 82 076001 (2019)
        <https://iopscience.iop.org/article/10.1088/1361-6633/ab1ca4>`_.

        2. P. E. M. F. Mendonça, R. d. J. Napolitano, M. A. Marchiolli, C. J. Foster, and
        Y.-C. Liang. *Alternative ﬁdelity measure between quantum states*,
        `Phys. Rev. A, 78 052330 (2008) <https://doi.org/10.1103/PhysRevA.78.052330>`_.

        3. A. Miszczak, Z. Puchala, P. Horodecki, A. Uhlmann, and K. Zyczkowski.
        *Sub- and super-ﬁdelity as bounds for quantum ﬁdelity*.
        `Quantum Inf. Comput., 9(1):103–130, (2009)
        <https://dl.acm.org/doi/10.5555/2021256.2021263>`_.
    """

    backend = _check_backend(backend)

    state = backend.cast(state, dtype=state.dtype)
    target = backend.cast(target, dtype=target.dtype)

    purity_state = purity(state, backend=backend)
    purity_target = purity(target, backend=backend)

    if (
        backend.np.abs(purity_state - 1) <= PRECISION_TOL
        or backend.np.abs(purity_target - 1) <= PRECISION_TOL
    ):
        return fidelity(state, target, backend=backend)

    fid = backend.np.trace(state @ target)
    fid += backend.np.sqrt(1 - purity_state) * backend.np.sqrt(1 - purity_target)

    return backend.np.real(fid)


def chen_fidelity(state, target, backend=None):
    """Return the Chen fidelity between two quantum states.

    For a quantum ``state`` :math:`\\rho` and a ``target`` quantum state :math:`\\sigma`,
    the Chen fidelity is defined as:

    .. math::
        F_{\\text{C}}(\\rho, \\, \\sigma) = \\frac{1 - r}{2} + \\frac{1 + r}{2} \\,
            F_{\\text{N}}(\\rho, \\, \\sigma)  \\, ,

    where :math:`\\text{tr}(\\varrho^{2})` is the :class:`qibo.quantum_info.purity` of
    a quantum state :math:`\\varrho`.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Chen fidelity between :math:`\\rho` and :math:`\\sigma`.

    References:
        1. Y.-C. Liang, Y.-H. Yeh, P. E. M. F. Mendonça, R. Y. Teh, M. D. Reid, and P. D. Drummond,
        *Quantum fidelity measures for mixed states*, `Rep. Prog. Phys. 82 076001 (2019)
        <https://iopscience.iop.org/article/10.1088/1361-6633/ab1ca4>`_.

        2. J.-L. Chen, L. Fu, A. A. Ungar, and X.-G. Zhao, *Alternative fidelity measure between
        two states of an N-state quantum system*, `Phys. Rev. A 65, 054304 (2002)
        <https://doi.org/10.1103/PhysRevA.65.054304>`_.

    """

    backend = _check_backend(backend)

    state = backend.cast(state, dtype=state.dtype)
    target = backend.cast(target, dtype=target.dtype)

    dims = len(state)

    coeff = 1 / (dims - 1)

    fid = (1 - coeff) / 2
    fid += ((1 + coeff) / 2) * n_fidelity(state, target, backend=backend)

    return fid


def geometric_mean_fidelity(state, target, backend=None):
    """Return the geometric-mean fidelity between two quantum states.

    For a quantum ``state`` :math:`\\rho` and a ``target`` quantum state :math:`\\sigma`,
    the geometric-mean fidelity is defined as:

    .. math::
        F_{\\text{GM}}(\\rho, \\, \\sigma) = \\frac{\\text{tr}(\\rho \\, \\sigma)}
            {\\sqrt{\\text{tr}(\\rho^{2}) \\, \\text{tr}(\\sigma^{2})}} \\, ,

    where :math:`\\text{tr}(\\varrho^{2})` is the :class:`qibo.quantum_info.purity` of
    a quantum state :math:`\\varrho`. If at least one of the quantum states is pure,
    then the geometric-mean fidelity reduces to the usual :class:`qibo.quantum_info.fidelity`.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Geometric-mean fidelity between :math:`\\rho` and :math:`\\sigma`.

    References:
        1. Y.-C. Liang, Y.-H. Yeh, P. E. M. F. Mendonça, R. Y. Teh, M. D. Reid, and P. D. Drummond,
        *Quantum fidelity measures for mixed states*, `Rep. Prog. Phys. 82 076001 (2019)
        <https://iopscience.iop.org/article/10.1088/1361-6633/ab1ca4>`_.

        2. X. Wang, C.-S. Yu, and X.X. Yi. *An alternative quantum ﬁdelity for mixed states
        of qudits*. `Phys. Lett. A, 373(1):58–60, (2008)
        <https://doi.org/10.1016/j.physleta.2008.10.083>`_.
    """

    backend = _check_backend(backend)

    state = backend.cast(state, dtype=state.dtype)
    target = backend.cast(target, dtype=target.dtype)

    purity_state = purity(state, backend=backend)
    purity_target = purity(target, backend=backend)

    if (
        backend.np.abs(purity_state - 1) <= PRECISION_TOL
        or backend.np.abs(purity_target - 1) <= PRECISION_TOL
    ):
        return fidelity(state, target, backend=backend)

    gm_fid = backend.np.trace(state @ target)
    gm_fid /= backend.np.sqrt(purity_state * purity_target)

    return backend.np.real(gm_fid)


def max_fidelity(state, target, backend=None):
    """Return max fidelity between two quantum states.

    For a quantum ``state`` :math:`\\rho` and a ``target`` quantum state :math:`\\sigma`,
    the max fidelity is defined as:

    .. math::
        F_{\\text{max}}(\\rho, \\, \\sigma) = \\frac{\\text{tr}(\\rho \\, \\sigma)}
            {\\text{max}(\\text{tr}(\\rho^{2}), \\, \\text{tr}(\\sigma^{2}))} \\, ,

    where :math:`\\text{tr}(\\varrho^{2})` is the :class:`qibo.quantum_info.purity` of
    a quantum state :math:`\\varrho`. If at least one of the quantum states is pure,
    then the max fidelity reduces to the usual :class:`qibo.quantum_info.fidelity`.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Max fidelity between :math:`\\rho` and :math:`\\sigma`.

    References:
        1. Y.-C. Liang, Y.-H. Yeh, P. E. M. F. Mendonça, R. Y. Teh, M. D. Reid, and P. D. Drummond,
        *Quantum fidelity measures for mixed states*, `Rep. Prog. Phys. 82 076001 (2019)
        <https://iopscience.iop.org/article/10.1088/1361-6633/ab1ca4>`_.
    """

    backend = _check_backend(backend)

    state = backend.cast(state, dtype=state.dtype)
    target = backend.cast(target, dtype=target.dtype)

    purity_state = purity(state, backend=backend)
    purity_target = purity(target, backend=backend)

    if (
        backend.np.abs(purity_state - 1) <= PRECISION_TOL
        or backend.np.abs(purity_target - 1) <= PRECISION_TOL
    ):
        return fidelity(state, target, backend=backend)

    max_fid = backend.np.trace(state @ target)
    max_fid /= max(purity_state, purity_target)

    return max_fid


def bures_angle(state, target, backend=None):
    """Calculates the Bures angle :math:`D_{A}` between a ``state``
    :math:`\\rho` and a ``target`` state :math:`\\sigma`. This is given by

    .. math::
        D_{A}(\\rho, \\, \\sigma) = \\text{arccos}\\left(\\sqrt{F(\\rho, \\, \\sigma)}\\right) \\, ,

    where :math:`F(\\rho, \\sigma)` is the :func:`qibo.quantum_info.fidelity`
    between `state` and `target`.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Bures angle between ``state`` and ``target``.
    """
    backend = _check_backend(backend)

    angle = backend.np.arccos(backend.np.sqrt(fidelity(state, target, backend=backend)))

    return angle


def bures_distance(state, target, backend=None):
    """Calculates the Bures distance :math:`D_{B}` between a ``state``
    :math:`\\rho` and a ``target`` state :math:`\\sigma`. This is given by

    .. math::
        D_{B}(\\rho, \\, \\sigma) = \\sqrt{2 \\, \\left(1 - \\sqrt{F(\\rho, \\, \\sigma)}\\right)}

    where :math:`F(\\rho, \\sigma)` is the :func:`qibo.quantum_info.fidelity`
    between `state` and `target`.

    Args:
        state (ndarray): statevector or density matrix.
        target (ndarray): statevector or density matrix.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Bures distance between ``state`` and ``target``.
    """
    backend = _check_backend(backend)

    sqrt_fid = backend.np.sqrt(fidelity(state, target, backend=backend))
    distance = backend.np.sqrt(2 * (1 - sqrt_fid))

    return distance


def process_fidelity(channel, target=None, check_unitary: bool = False, backend=None):
    """Process fidelity between a quantum ``channel`` :math:`\\mathcal{E}` and
    a ``target`` unitary channel :math:`U`. The process fidelity is defined as

    .. math::
        F_{\\text{pro}}(\\mathcal{E}, \\mathcal{U}) = \\frac{1}{d^{2}} \\,
            \\text{tr}(\\mathcal{E}^{\\dagger} \\, \\mathcal{U})

    Args:
        channel: quantum channel :math:`\\mathcal{E}`.
        target (optional): quantum channel :math:`U`. If ``None``, target is the
            Identity channel. Defaults to ``None``.
        check_unitary (bool, optional): if ``True``, checks if one of the
            input channels is unitary. Default: ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Process fidelity between ``channel`` and ``target``.
    """
    backend = _check_backend(backend)

    if target is not None:
        if channel.shape != target.shape:
            raise_error(
                TypeError,
                f"Channels must have the same dims, but {channel.shape} != {target.shape}",
            )

    dim = int(np.sqrt(channel.shape[0]))

    if check_unitary is True:
        norm_channel = float(
            backend.calculate_matrix_norm(
                backend.np.matmul(
                    backend.np.conj(backend.np.transpose(channel, (1, 0))), channel
                )
                - backend.np.eye(dim**2)
            )
        )
        if target is None and norm_channel > PRECISION_TOL:
            raise_error(TypeError, "Channel is not unitary and Target is None.")
        if target is not None:
            norm_target = float(
                backend.calculate_vector_norm(
                    backend.np.matmul(
                        backend.np.conj(backend.np.transpose(target, (1, 0))), target
                    )
                    - backend.np.eye(dim**2)
                )
            )
            if (norm_channel > PRECISION_TOL) and (norm_target > PRECISION_TOL):
                raise_error(TypeError, "Neither channel is unitary.")

    if target is None:
        # With no target, return process fidelity with Identity channel
        process_fid = backend.np.real(backend.np.trace(channel)) / dim**2
        process_fid = float(process_fid)

        return process_fid

    process_fid = backend.np.matmul(
        backend.np.transpose(backend.np.conj(channel), (1, 0)), target
    )
    process_fid = backend.np.real(backend.np.trace(process_fid)) / dim**2

    return process_fid


def process_infidelity(channel, target=None, check_unitary: bool = False, backend=None):
    """Process infidelity between quantum channel :math:`\\mathcal{E}` and a
    ``target`` unitary channel :math:`U`. The process infidelity is defined as

    .. math::
        1 - F_{\\text{pro}}(\\mathcal{E}, \\mathcal{U}) \\, ,

    where :math:`F_{\\text{pro}}` is the :func:`qibo.quantum_info.process_fidelity`.

    Args:
        channel: quantum channel :math:`\\mathcal{E}`.
        target (optional): quantum channel :math:`U`. If ``None``, target is the
            Identity channel. Defaults to ``None``.
        check_unitary (bool, optional): if ``True``, checks if one of the
            input channels is unitary. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.


    Returns:
        float: Process infidelity between ``channel`` :math:`\\mathcal{E}`
        and ``target`` :math:`U`.
    """
    return 1 - process_fidelity(
        channel, target=target, check_unitary=check_unitary, backend=backend
    )


def average_gate_fidelity(
    channel, target=None, check_unitary: bool = False, backend=None
):
    """Average gate fidelity between a quantum ``channel`` :math:`\\mathcal{E}`
    and a ``target`` unitary channel :math:`U`. The average gate fidelity is
    defined as

    .. math::
        F_{\\text{avg}}(\\mathcal{E}, \\mathcal{U}) = \\frac{d \\,
            F_{pro}(\\mathcal{E}, \\mathcal{U}) + 1}{d + 1}

    where :math:`d` is the dimension of the channels and
    :math:`F_{pro}(\\mathcal{E}, \\mathcal{U})` is the
    :meth:`~qibo.metrics.process_fidelily` of channel
    :math:`\\mathcal{E}` with respect to the unitary
    channel :math:`\\mathcal{U}`.

    Args:
        channel: quantum channel :math:`\\mathcal{E}`.
        target (optional): quantum channel :math:`\\mathcal{U}`.
            If ``None``, target is the Identity channel. Defaults to ``None``.
        check_unitary (bool, optional): if ``True``, checks if one of the
            input channels is unitary. Default: ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Process fidelity between channel :math:`\\mathcal{E}`
        and target unitary channel :math:`\\mathcal{U}`.
    """

    dim = channel.shape[0]

    process_fid = process_fidelity(
        channel, target, check_unitary=check_unitary, backend=backend
    )
    process_fid = (dim * process_fid + 1) / (dim + 1)

    return process_fid


def gate_error(channel, target=None, check_unitary: bool = False, backend=None):
    """Gate error between a quantum ``channel`` :math:`\\mathcal{E}` and a
    ``target`` unitary channel :math:`U`, which is defined as

    .. math::
        E(\\mathcal{E}, \\mathcal{U}) = 1 - F_{\\text{avg}}(\\mathcal{E}, \\mathcal{U}) \\, ,

    where :math:`F_{\\text{avg}}(\\mathcal{E}, \\mathcal{U})` is the
    :func:`qibo.quantum_info.average_gate_fidelity`.

    Args:
        channel: quantum channel :math:`\\mathcal{E}`.
        target (optional): quantum channel :math:`\\mathcal{U}`. If ``None``,
            target is the Identity channel. Defaults to ``None``.
        check_unitary (bool, optional): if ``True``, checks if one of the
            input channels is unitary. Default: ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Gate error between ``channel`` :math:`\\mathcal{E}`
        and ``target`` :math:`\\mathcal{U}`.
    """
    error = 1 - average_gate_fidelity(
        channel, target, check_unitary=check_unitary, backend=backend
    )

    return error


def diamond_norm(channel, target=None, backend=None, **kwargs):  # pragma: no cover
    """Calculates the diamond norm :math:`\\|\\mathcal{E}\\|_{\\diamond}` of
    ``channel`` :math:`\\mathcal{E}`, which is given by

    .. math::
        \\|\\mathcal{E}\\|_{\\diamond} = \\max_{\\rho} \\, \\| \\left(\\mathcal{E} \\otimes I_{d^{2}}\\right)(\\rho) \\|_{1} \\, ,

    where :math:`I_{d^{2}}` is the :math:`d^{2} \\times d^{2}` Identity operator,
    :math:`d = 2^{n}`, :math:`n` is the number of qubits,
    and :math:`\\|\\cdot\\|_{1}` denotes the trace norm.

    If a ``target`` channel :math:`\\Lambda` is specified,
    then it calculates :math:`\\| \\mathcal{E} - \\Lambda\\|_{\\diamond}`.

    Example::

        from qibo.quantum_info import diamond_norm, random_unitary, to_choi

        nqubits = 1
        dim = 2**nqubits

        unitary = random_unitary(dim)
        unitary = to_choi(unitary, order="row")

        unitary_2 = random_unitary(dim)
        unitary_2 = to_choi(unitary_2, order="row")

        dnorm = diamond_norm(unitary, unitary_2)

    Args:
        channel (ndarray): row-vectorized Choi representation of a quantum channel.
        target (ndarray, optional): row-vectorized Choi representation of a target
            quantum channel. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.
        kwargs: optional arguments to pass to CVXPY solver. For more information,
            please visit `CVXPY's API documentation
            <https://www.cvxpy.org/api_reference/cvxpy.problems.html#problem>`_.

    Returns:
        float: diamond norm of either ``channel`` or ``channel - target``.

    .. note::
        This function requires the optional CVXPY package to be installed.
    """
    import cvxpy  # pylint: disable=import-outside-toplevel  #type: ignore

    backend = _check_backend(backend)

    if target is not None:
        if channel.shape != target.shape:
            raise_error(
                TypeError,
                f"Channels must have the same dims, but {channel.shape} != {target.shape}",
            )

    if target is not None:
        channel -= target

    # `CVXPY` only works with `numpy`, so this function has to
    # convert any channel to the `numpy` backend by default
    backend = _check_backend(backend)

    channel = backend.to_numpy(channel)

    channel = backend.np.transpose(channel, (1, 0))
    channel_real = backend.np.real(channel)
    channel_imag = backend.np.imag(channel)

    dim = int(np.sqrt(channel.shape[0]))

    first_variables_real = cvxpy.Variable(shape=(dim, dim))
    first_variables_imag = cvxpy.Variable(shape=(dim, dim))
    first_variables = cvxpy.bmat(
        [
            [first_variables_real, -first_variables_imag],
            [first_variables_imag, first_variables_real],
        ]
    )

    second_variables_real = cvxpy.Variable(shape=(dim, dim))
    second_variables_imag = cvxpy.Variable(shape=(dim, dim))
    second_variables = cvxpy.bmat(
        [
            [second_variables_real, -second_variables_imag],
            [second_variables_imag, second_variables_real],
        ]
    )

    variables_real = cvxpy.Variable(shape=(dim**2, dim**2))
    variables_imag = cvxpy.Variable(shape=(dim**2, dim**2))
    identity = sparse.eye(dim)

    constraints_real = cvxpy.bmat(
        [
            [cvxpy.kron(identity, first_variables_real), variables_real],
            [variables_real.T, cvxpy.kron(identity, second_variables_real)],
        ]
    )
    constraints_imag = cvxpy.bmat(
        [
            [cvxpy.kron(identity, first_variables_imag), variables_imag],
            [-variables_imag.T, cvxpy.kron(identity, second_variables_imag)],
        ]
    )
    constraints_block = cvxpy.bmat(
        [[constraints_real, -constraints_imag], [constraints_imag, constraints_real]]
    )

    constraints = [
        first_variables >> 0,
        first_variables_real == first_variables_real.T,
        first_variables_imag == -first_variables_imag.T,
        cvxpy.trace(first_variables_real) == 1,
        second_variables >> 0,
        second_variables_real == second_variables_real.T,
        second_variables_imag == -second_variables_imag.T,
        cvxpy.trace(second_variables_real) == 1,
        constraints_block >> 0,
    ]

    objective_function = cvxpy.Maximize(
        cvxpy.trace(channel_real @ variables_real)
        + cvxpy.trace(channel_imag @ variables_imag)
    )
    problem = cvxpy.Problem(objective=objective_function, constraints=constraints)
    solution = problem.solve(**kwargs)

    return solution


def expressibility(
    circuit,
    power_t: int,
    samples: int,
    order: Optional[Union[int, float, str]] = 2,
    backend=None,
):
    """Returns the expressibility :math:`\\|A\\|` of a parametrized circuit,
    where

    .. math::
        A = \\int_{\\text{Haar}} d\\psi \\, \\left(|\\psi\\rangle\\right.\\left.
            \\langle\\psi|\\right)^{\\otimes t} - \\int_{\\Theta} d\\psi \\,
            \\left(|\\psi_{\\theta}\\rangle\\right.\\left.
            \\langle\\psi_{\\theta}|\\right)^{\\otimes t}

    Args:
        circuit (:class:`qibo.models.Circuit`): Parametrized circuit.
        power_t (int): power that defines the :math:`t`-design.
        samples (int): number of samples to estimate the integrals.
        order (int or float or str, optional): order of the norm :math:`\\|A\\|`.
            For specifications, see :meth:`qibo.backends.abstract.calculate_norm`.
            Defaults to :math:`2`.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Expressibility of parametrized circuit.
    """

    if isinstance(power_t, int) is False:
        raise_error(
            TypeError, f"power_t must be type int, but it is type {type(power_t)}."
        )

    if isinstance(samples, int) is False:
        raise_error(
            TypeError, f"samples must be type int, but it is type {type(samples)}."
        )

    from qibo.quantum_info.utils import (  # pylint: disable=C0415
        haar_integral,
        pqc_integral,
    )

    backend = _check_backend(backend)

    deviation = haar_integral(
        circuit.nqubits, power_t, samples=None, backend=backend
    ) - pqc_integral(circuit, power_t, samples, backend=backend)

    fid = float(backend.calculate_vector_norm(deviation, order=order))

    return fid


def frame_potential(
    circuit,
    power_t: int,
    samples: int = None,
    backend=None,
):
    """Returns the frame potential of a parametrized circuit under uniform
    sampling of the parameters.

    For :math:`n` qubits and moment :math:`t`, the frame potential
    :math:`\\mathcal{F}_{\\mathcal{U}}^{(t)}` if given by [1]

    .. math::
        \\mathcal{F}_{\\mathcal{U}}^{(t)} = \\int_{U,V \\in \\mathcal{U}} \\,
            \\text{d}U \\, \\text{d}V \\, \\bigl| \\, \\text{tr}(U^{\\dagger} \\, V)
            \\, \\bigr|^{2t} \\, ,

    where :math:`\\mathcal{U}` is the group of unitaries defined by the parametrized circuit.
    The frame potential is approximated by the average

    .. math::
        \\mathcal{F}_{\\mathcal{U}}^{(t)} \\approx \\frac{1}{N} \\,
            \\sum_{k=1}^{N} \\, \\bigl| \\, \\text{tr}(U_{k}^{\\dagger} \\, V_{k}) \\, \\bigr|^{2t} \\, ,

    where :math:`N` is the number of ``samples``.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): Parametrized circuit.
        power_t (int): power that defines the :math:`t`-design.
        samples (int): number of samples to estimate the integral.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        float: Frame potential of the parametrized circuit.

    References:
        1. M. Liu *et al.*, *Estimating the randomness of quantum circuit ensembles up to 50 qubits*.
        `arXiv:2205.09900 [quant-ph] <https://arxiv.org/abs/2205.09900>`_.
    """
    if not isinstance(power_t, int):
        raise_error(
            TypeError, f"power_t must be type int, but it is type {type(power_t)}."
        )

    if not isinstance(samples, int):
        raise_error(
            TypeError, f"samples must be type int, but it is type {type(samples)}."
        )

    backend = _check_backend(backend)

    nqubits = circuit.nqubits
    dim = 2**nqubits

    potential = 0
    for _ in range(samples):
        unitary_1 = circuit.copy()
        params_1 = np.random.uniform(-np.pi, np.pi, circuit.trainable_gates.nparams)
        unitary_1.set_parameters(params_1)
        unitary_1 = unitary_1.unitary(backend) / np.sqrt(dim)

        for _ in range(samples):
            unitary_2 = circuit.copy()
            params_2 = np.random.uniform(-np.pi, np.pi, circuit.trainable_gates.nparams)
            unitary_2.set_parameters(params_2)
            unitary_2 = unitary_2.unitary(backend) / np.sqrt(dim)

            potential += backend.np.abs(
                backend.np.trace(
                    backend.np.transpose(backend.np.conj(unitary_1), (1, 0)) @ unitary_2
                )
            ) ** (2 * power_t)

    return potential / samples**2


def quantum_fisher_information_matrix(
    circuit,
    parameters=None,
    initial_state=None,
    return_complex: bool = True,
    backend=None,
):
    """Calculate the Quantum Fisher Information Matrix (QFIM) of a parametrized ``circuit``.

    Given a set of ``parameters`` :math:`\\theta = \\{\\theta_{k}\\}_{k\\in[M]}` and a
    parameterized unitary ``circuit`` :math:`U(\\theta)` acting on an ``initial_state``
    :math:`\\ket{\\phi}`, the QFIM is such that its elements can be calculated as

    .. math::
        \\mathbf{F}_{jk} = 4 \\, \\text{Re}\\left\\{ \\braket{\\partial_{j} \\psi | \\partial_{k}
            \\psi} - \\braket{\\partial_{j} \\psi | \\psi}\\!\\braket{\\psi | \\partial_{k} \\psi}
            \\right\\} \\, ,

    where we have used the short notations :math:`\\ket{\\psi} \\equiv \\ket{\\psi(\\theta)}
    = U(\\theta) \\ket{\\phi}`, and :math:`\\ket{\\partial_{k} \\psi} \\equiv \\frac{\\partial}
    {\\partial\\theta_{k}} \\ket{\\psi(\\theta)}`.
    If the ``initial_state`` :math:`\\ket{\\phi}` is not specified, it defaults to
    :math:`\\ket{0}^{\\otimes n}`.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): parametrized circuit :math:`U(\\theta)`.
        parameters (ndarray, optional): parameters whose QFIM to calculate.
            If ``None``, QFIM is calculated with the paremeters from ``circuit``, i.e.
            ``parameters = circuit.get_parameters()``. Defaults to ``None``.
        initial_state (ndarray, optional): Initial configuration. It can be specified
            by the setting the state vector using an array or a circuit. If ``None``,
            the initial state is :math:`\\ket{0}^{\\otimes n}`. Defaults to ``None``.
        return_complex (bool, optional): If ``True``, calculates the Jacobian matrix
            of real and imaginary parts of :math:`\\ket{\\psi(\\theta)}`. If ``False``,
            calculates only the Jacobian matrix of the real part. Defaults to ``True``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: Quantum Fisher Information :math:`\\mathbf{F}`.
    """
    backend = _check_backend(backend)

    if parameters is None:
        parameters = circuit.get_parameters()
        parameters = backend.cast(parameters, dtype=float).flatten()

    jacobian = backend.calculate_jacobian_matrix(
        circuit, parameters, initial_state, return_complex
    )

    if return_complex:
        jacobian = jacobian[0] + 1j * jacobian[1]

    jacobian = backend.cast(jacobian, dtype=np.complex128)

    copied = circuit.copy(deep=True)
    copied.set_parameters(parameters)

    state = backend.execute_circuit(copied, initial_state=initial_state).state()

    overlaps = jacobian.T @ state

    qfim = jacobian.T @ jacobian
    qfim = qfim - backend.np.outer(overlaps, backend.np.conj(overlaps.T))

    return 4 * backend.np.real(qfim)
